// Bullet Hell by SkyBerron <skyberron@gmail.com>
// This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International License.
// To view a copy of the license, visit https://creativecommons.org/licenses/by-nc-nd/4.0/legalcode

const minTX = 0, maxTX = 28;
const minTY = 0, maxTY = 21;
const minX = ( ( minTX ) << 8 ), maxX = ( ( maxTX ) << 8 );
const minY = ( ( minTY ) << 8 ), maxY = ( ( maxTY ) << 8 );
const lvl = builtin("sLvl");
const pts = builtin("sPts");
const player = builtin("helmet8");
const side = builtin("helmet2");
const enemy = builtin("char26");
const star = builtin("sStar");
const weapon = builtin("sHeart");
const bullet = builtin("sBtn");
const bgColor = 225;
const playerColor = 7;
const enemyColor = 40;
const weaponColor = 80;
const bulletColor = 200;
const starColor = 79;


const cWeapon = 16 << 2;
var nWeapon;
var tWeapon = Array( cWeapon );
const vWeapon = 3 << 5;
var timWeapon;
const timMaxWeapon = 5;

const cBullet = 100 << 2;
var nBullet;
var tBullet = Array( cBullet );
const vBulletMin = 1 << 5;
const vBulletMax = 1 << 7;

const cEnemy = 10 << 4;
var nEnemy;
var tEnemy = Array( cEnemy );

var dead;
var score;
var frame;
var bx, by, bvx, bvy;


tileshift( 0, 0 );
fill( bgColor );

reset();


/*
// Linear interpolation
function lerp( a, b, c, d ){
    return( a + ( ( ( b - a ) * c ) / d ) );
}
*/

function reset(){
    highscore( score );
    score = 0;
    
    nWeapon = 0;
    nBullet = 0;
    timWeapon = timMaxWeapon;
    
    bx = ( minX + maxX ) >> 1;
    by = maxY - ( 32 << 5 );
    
    bvx = 2 << 5;
    bvy = bvx;
    dead = false;
    
    frame = 0;
    
    yEnemy = minY - ( 8 << 5 );
    nEnemy = cEnemy;
    var i = 0;
    for( var iEnemy = 0; iEnemy < nEnemy; iEnemy += 16 ) {
        // [0] = xc
        // [1] = yc
        // [2] = ax
        // [3] = wx
        // [4] = px
        // [5] = vy
        // [6] = frames
        // [7] = vbullet
        // [8] = nbullet
        // [9] = pbullet
        // [10] = ipbullet
        // [11] = life
        
        tEnemy[ iEnemy ] = ( minX + maxX ) >> 1;
        tEnemy[ iEnemy + 1 ] = yEnemy;
        tEnemy[ iEnemy + 2 ] = ( ( maxX - minX ) >> 1 ) - ( 16 << 5 );
        //tEnemy[ iEnemy + 3 ] = lerp( 1, 6, iEnemy, nEnemy );
        tEnemy[ iEnemy + 3 ] = i;
        //tEnemy[ iEnemy + 4 ] = 0;
        tEnemy[ iEnemy + 5 ] = 1 << 2;
        //tEnemy[ iEnemy + 6 ] = lerp( 32, 16, iEnemy, nEnemy );
        tEnemy[ iEnemy + 6 ] = 32 - ( i << 1 );
        //tEnemy[ iEnemy + 7 ] = lerp( vBulletMin, vBulletMax, iEnemy, nEnemy );
        tEnemy[ iEnemy + 7 ] = vBulletMin + ( i << 4 );
        //tEnemy[ iEnemy + 8 ] = 8;
        tEnemy[ iEnemy + 9 ] = 0;
        tEnemy[ iEnemy + 10 ] = 1 << 5;
        tEnemy[ iEnemy + 11 ] = 64;
        
        yEnemy -= 64 << 5;
        i++;
    }
}

// Rect overlap test
function test_collide( x1, y1, x2, y2, ex, ey ){
    var dx = x2 - x1, dy = y2 - y1;
    return( ( dx > -ex ) && ( dx < ex ) && ( dy > -ey ) && ( dy < ey ) );
}

/*
// Circle overlap test
function test_collide2( x1, y1, x2, y2, rx, ry ){
    var dx = x2 - x1, dy = y2 - y1, d = rx + ry;
    return( ( ( dx * dx ) + ( dy * dy ) ) < ( d * d ) );
}
*/

function test_visible( x, y ){
    return( ( x >= minX ) && ( x <= maxX ) && ( y >= minY ) && ( y <= maxY ) );
}


function updateEnemy(iEnemy){
    // [0] = xc
    // [1] = yc
    // [2] = ax
    // [3] = wx
    // [4] = px
    // [5] = vy
    // [6] = frames
    // [7] = vbullet
    // [8] = nbullet
    // [9] = pbullet
    // [10] = ipbullet
    // [11] = life
    var i, k;

    var y = tEnemy[iEnemy + 1];
    //var x = tEnemy[iEnemy] + ( ( tEnemy[iEnemy + 2] * sin( tEnemy[iEnemy + 4] + ( tEnemy[iEnemy + 3] * frame ) ) ) >> 8 );
    var x = tEnemy[iEnemy] + ( ( tEnemy[iEnemy + 2] * sin( ( tEnemy[iEnemy + 3] * frame ) ) ) >> 8 );
    sprite( ( x >> 5 ) - 8, ( y >> 5 ) - 8, enemy );
    
    //if( ( x >= minX ) && ( x <= maxX ) && ( y >= minY ) && ( y <= maxY ) ) {
    if( test_visible( x, y ) ) {
        if( ( frame % tEnemy[ iEnemy + 6 ] ) == 0 ) {
            //var n = tEnemy[ iEnemy + 8 ];
            var n = 8;
            var ang0 = tEnemy[ iEnemy + 9 ];
            tEnemy[ iEnemy + 9 ] = ang0 + tEnemy[ iEnemy + 10 ];
            //for( i = 0; ( i < n ) && ( iBullet < cBullet ); i++ ) {
            for( i = 0; i < n; i++ ) {
                var v = tEnemy[ iEnemy + 7 ];
                tBullet[ nBullet ] = x;
                tBullet[ nBullet + 1 ] = y;
                var ang = ang0 + ( ( i << 10 ) / n );
                tBullet[ nBullet + 2 ] = ( v * cos( ang ) ) >> 8;
                tBullet[ nBullet + 3 ] = ( v * sin( ang ) ) >> 8;
                nBullet += 4;
            }
        }
    }
    for( iWeapon = 0; iWeapon < nWeapon; iWeapon += 4 ) {
        var xw = tWeapon[iWeapon];
        var yw = tWeapon[iWeapon + 1];
        if( test_collide( x, y, xw, yw, ( 11 << 5 ), ( 11 << 5 ) ) ) {
            tEnemy[iEnemy + 11]--;
            tWeapon[iWeapon + 3] = 0;
        }
    }
    /*
    if( test_collide( x, y, bx, by, ( 14 << 5 ), ( 14 << 5 ) ) ) {
        dead = true;
    }
    */
    tEnemy[iEnemy + 1] = y + tEnemy[iEnemy + 5];
    
    if( tEnemy[iEnemy + 11] <= 0 ) {
        //delete_enemy( iEnemy );
        for( k = iEnemy + 16; k < nEnemy; k++ ) {
            tEnemy[k - 16] = tEnemy[k];
        }
        nEnemy -= 16;
        score++;
    } else {
        iEnemy += 16;
    }
    
    //iEnemy += 16;
    
    return(iEnemy);
}

function update(){
    var iEnemy, iBullet, iWeapon, k;
    var x, y, v;
    
    frame++;
    
    color( enemyColor );
    for( iEnemy = 0; iEnemy < nEnemy; ) {
        iEnemy = updateEnemy(iEnemy);
    }

    color( bulletColor );
    for( iBullet = 0; iBullet < nBullet; ) {
        x = tBullet[iBullet];
        y = tBullet[iBullet + 1];
        //if( ( x >= minX ) && ( x <= maxX ) && ( y >= minY ) && ( y <= maxY ) ) {
        if( test_visible( x, y ) ) {
            sprite( ( x >> 5 ) - 4, ( y >> 5 ) - 4, bullet );
            tBullet[iBullet] = x + tBullet[iBullet + 2];
            tBullet[iBullet + 1] = y + tBullet[iBullet + 3];
            iBullet += 4;
        } else {
            for( k = iBullet + 4; k < nBullet; k++ ) {
                tBullet[k - 4] = tBullet[k];
            }
            nBullet -= 4;
        }
        if( test_collide( x, y, bx, by, ( 6 << 5 ), ( 6 << 5 ) ) ) {
            dead = true;
        }
    }
    
    color( weaponColor );
    for( iWeapon = 0; iWeapon < nWeapon; ) {
        x = tWeapon[iWeapon];
        y = tWeapon[iWeapon + 1];
        sprite( ( x >> 5 ) - 4, ( y >> 5 ) - 4, weapon );
        
        if( tWeapon[iWeapon + 3] <= 0 ) {
            //delete_weapon( iWeapon );
            for( k = iWeapon + 4; k < nWeapon; k++ ) {
                tWeapon[k - 4] = tWeapon[k];
            }
            nWeapon -= 4;
        } else {
            var wv = tWeapon[iWeapon + 2];
            tWeapon[iWeapon + 1] = y - wv;
            tWeapon[iWeapon + 3] -= wv;
            iWeapon += 4;
        }
    }
    
    if( pressed( "C" ) ) {
        exit();
    }
    
    v = minX + ( 8 << 5 );
    if( pressed( "LEFT" ) ) {
        bx -= bvx;
        if( bx < v ){
            bx = v;
        }
    }
    v = maxX - ( 8 << 5 );
    if( pressed( "RIGHT" ) ) {
        bx += bvx;
        if( bx > v ) {
            bx = v;
        }
    }
    v = minY + ( 8 << 5 );
    if( pressed( "UP" ) ) {
        by -= bvy;
        if( by < v ){
            by = v;
        }
    }
    v = maxY - ( 8 << 5 );
    if( pressed( "DOWN" ) ) {
        by += bvy;
        if( by > v ){
            by = v;
        }
    }
    
    timWeapon--;
    if( justPressed( "A" ) && ( timWeapon <= 0 ) ){
        //add_weapon( bx, by - ( 8 << 5 ), vWeapon, dWeapon );
        //if( nWeapon < cWeapon ) {
            tWeapon[nWeapon] = bx;
            tWeapon[nWeapon + 1] = by - ( 8 << 5 );
            tWeapon[nWeapon + 2] = vWeapon;
            tWeapon[nWeapon + 3] = maxX;
            nWeapon += 4;
            timWeapon = timMaxWeapon;
        //}
    }
    
    color( playerColor );
    sprite( ( ( bx ) >> 5 ) - 8, ( ( by ) >> 5 ) - 8, player );

    // hud
    color( 47 );
    sprite( 100, 165, pts );
    cursor( 120, 165 );
    printNumber( score );
    
    if( dead ) {
        reset();
    }
}

